@@ -1,7 +1,7 @@
 use ark_ff::Field;
 use core::fmt::Debug;
 
-use crate::crh::{CRHScheme, TwoToOneCRHScheme};
+use crate::crh::{CRHScheme, TwoToOneCRHScheme, MMRTwoToOneCRHScheme};
 use ark_relations::r1cs::SynthesisError;
 
 use ark_r1cs_std::prelude::*;
@@ -50,7 +50,7 @@ pub trait TwoToOneCRHSchemeGadget<H: TwoToOneCRHScheme, ConstraintF: Field>: Siz
     ) -> Result<Self::OutputVar, SynthesisError>;
 }
 
-pub trait MMRTwoToOneCRHSchemeGadget<H: TwoToOneCRHScheme, ConstraintF: Field>: Sized {
+pub trait MMRTwoToOneCRHSchemeGadget<H: MMRTwoToOneCRHScheme, ConstraintF: Field>: Sized {
     type InputVar: ?Sized;
     type OutputVar: EqGadget<ConstraintF>
         + ToBytesGadget<ConstraintF>
@@ -76,15 +76,15 @@ pub trait MMRTwoToOneCRHSchemeGadget<H: TwoToOneCRHScheme, ConstraintF: Field>:
     ) -> Result<Self::OutputVar, SynthesisError>;
 
     fn left_compress( 
-        parameters: &Self::Parameters,
-        left_input: &Self::Output,
-        right_input: &Self::Input,
-    ) -> Result<Self::Output, Error>;
+        parameters: &Self::ParametersVar,
+        left_input: &Self::OutputVar,
+        right_input: &Self::InputVar,
+    ) -> Result<Self::OutputVar, SynthesisError>;
 
 
     fn right_compress(
-        parameters: &Self::Parameters,
-        left_input: &Self::Input,
-        right_input: &Self::Output,
-    ) -> Result<Self::Output, Error>;
+        parameters: &Self::ParametersVar,
+        left_input: &Self::InputVar,
+        right_input: &Self::OutputVar,
+    ) -> Result<Self::OutputVar, SynthesisError>;
 }
